@@ -58,6 +58,9 @@ static const char *rbac_role;
 /* --cid: Cluster ID for "join-cluster" command. */
 static struct uuid cid;
 
+/* --election-timer: Election timer for "create-cluster" command. */
+static uint64_t election_timer;
+
 static const struct ovs_cmdl_command *get_all_commands(void);
 
 OVS_NO_RETURN static void usage(void);
@@ -85,12 +88,14 @@ parse_options(int argc, char *argv[])
 {
     enum {
         OPT_RBAC_ROLE = UCHAR_MAX + 1,
-        OPT_CID
+        OPT_CID,
+        OPT_ELECTION_TIMER,
     };
     static const struct option long_options[] = {
         {"more", no_argument, NULL, 'm'},
         {"rbac-role", required_argument, NULL, OPT_RBAC_ROLE},
         {"cid", required_argument, NULL, OPT_CID},
+        {"election-timer", required_argument, NULL, OPT_ELECTION_TIMER},
         {"verbose", optional_argument, NULL, 'v'},
         {"help", no_argument, NULL, 'h'},
         {"option", no_argument, NULL, 'o'},
@@ -100,6 +105,7 @@ parse_options(int argc, char *argv[])
     char *short_options = ovs_cmdl_long_options_to_short_options(long_options);
 
     for (;;) {
+        struct ovsdb_error *error;
         int c;
 
         c = getopt_long(argc, argv, short_options, long_options, NULL);
@@ -122,6 +128,14 @@ parse_options(int argc, char *argv[])
             }
             break;
 
+        case OPT_ELECTION_TIMER:
+            election_timer = atoll(optarg);
+            error = raft_validate_election_timer(election_timer);
+            if (error) {
+                ovs_fatal(0, "%s", ovsdb_error_to_string_free(error));
+            }
+            break;
+
         case 'h':
             usage();
 
@@ -153,7 +167,7 @@ usage(void)
     printf("%s: Open vSwitch database management utility\n"
            "usage: %s [OPTIONS] COMMAND [ARG...]\n"
            "  create [DB [SCHEMA]]    create DB with the given SCHEMA\n"
-           "  create-cluster DB CONTENTS LOCAL\n"
+           "  [--election-timer=ms] create-cluster DB CONTENTS LOCAL\n"
            "    create clustered DB with given CONTENTS and LOCAL address\n"
            "  [--cid=UUID] join-cluster DB NAME LOCAL REMOTE...\n"
            "    join clustered DB with given NAME and LOCAL and REMOTE addrs\n"
@@ -303,7 +317,7 @@ do_create_cluster(struct ovs_cmdl_context *ctx)
     /* Create database file. */
     struct json *snapshot = json_array_create_2(schema_json, data);
     check_ovsdb_error(raft_create_cluster(db_file_name, schema->name,
-                                          local, snapshot));
+                                          local, snapshot, election_timer));
     ovsdb_schema_destroy(schema);
     json_destroy(snapshot);
 }
@@ -391,6 +405,9 @@ compact_or_convert(const char *src_name_, const char *dst_name_,
         ovs_fatal(retval, "%s: failed to lock lockfile", dst_name);
     }
 
+    /* Resulted DB will contain a single transaction without diff anyway. */
+    ovsdb_file_column_diff_disable();
+
     /* Save a copy. */
     struct ovsdb *ovsdb = (new_schema
                            ? ovsdb_file_read_as_schema(src_name, new_schema)
@@ -648,6 +665,8 @@ static void
 print_db_changes(struct shash *tables, struct smap *names,
                  const struct ovsdb_schema *schema)
 {
+    struct json *is_diff = shash_find_data(tables, "_is_diff");
+    bool diff = (is_diff && is_diff->type == JSON_TRUE);
     struct shash_node *n1;
 
     int i = 0;
@@ -691,7 +710,8 @@ print_db_changes(struct shash *tables, struct smap *names,
                     printf(" insert row %.8s:\n", row_uuid);
                 }
             } else {
-                printf(" row %s (%.8s):\n", old_name, row_uuid);
+                printf(" row %s (%.8s)%s:\n", old_name, row_uuid,
+                                              diff ? " diff" : "");
             }
 
             if (columns->type == JSON_OBJECT) {
@@ -1488,16 +1508,58 @@ do_check_cluster(struct ovs_cmdl_context *ctx)
                 if (!must_equal || raft_entry_equals(ae, be)) {
                     continue;
                 }
-                char *as = json_to_string(raft_entry_to_json(ae), JSSF_SORT);
-                char *bs = json_to_string(raft_entry_to_json(be), JSSF_SORT);
+                struct json *jae = raft_entry_to_json(ae);
+                struct json *jbe = raft_entry_to_json(be);
+                char *as = json_to_string(jae, JSSF_SORT);
+                char *bs = json_to_string(jbe, JSSF_SORT);
                 ovs_fatal(0, "log entries with index %"PRIu64" differ:\n"
                           "%s has %s\n"
                           "%s has %s",
                           idx, a->filename, as, b->filename, bs);
+                json_destroy(jae);
+                json_destroy(jbe);
             }
         }
     }
 
+    /* Check for db consistency:
+     * The serverid must be in the servers list.
+     */
+
+    for (struct server *s = c.servers; s < &c.servers[c.n_servers]; s++) {
+        struct shash *servers_obj = json_object(s->snap->servers);
+        char *server_id = xasprintf(SID_FMT, SID_ARGS(&s->header.sid));
+        bool found = false;
+        const struct shash_node *node;
+
+        SHASH_FOR_EACH (node, servers_obj) {
+            if (!strncmp(server_id, node->name, SID_LEN)) {
+                found = true;
+            }
+        }
+
+        if (!found) {
+            for (struct raft_entry *e = s->entries;
+                 e < &s->entries[s->log_end - s->log_start]; e++) {
+                if (e->servers == NULL) {
+                    continue;
+                }
+                struct shash *log_servers_obj = json_object(e->servers);
+                SHASH_FOR_EACH (node, log_servers_obj) {
+                    if (!strncmp(server_id, node->name, SID_LEN)) {
+                        found = true;
+                    }
+                }
+            }
+        }
+
+        if (!found) {
+            ovs_fatal(0, "%s: server %s not found in server list",
+                      s->filename, server_id);
+        }
+        free(server_id);
+    }
+
     /* Clean up. */
 
     for (size_t i = 0; i < c.n_servers; i++) {
